# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.3 $
from Products.ParsedXML.NodePath.ElementIdPath import ElementIdPathScheme
from WidgetDOM import getWidgetWrappedNode

class WidgetPathScheme(ElementIdPathScheme):
    def resolve_steps(self, top_node, steps):
        top_node = getWidgetWrappedNode(top_node)
        return ElementIdPathScheme.resolve_steps(
            self, top_node, steps)
    



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.3 $
import urllib

def add_and_edit(self, id, REQUEST):
    """Helper function to point to the object's management screen if
    'Add and Edit' button is pressed.
    id -- id of the object we just added
    """
    if REQUEST is None:
        return
    try:
        u = self.DestinationURL()
    except:
        u = REQUEST['URL1']
    if REQUEST.has_key('submit_edit'):
        u = "%s/%s" % (u, urllib.quote(id))
    REQUEST.RESPONSE.redirect(u+'/manage_main')



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.14 $
from Products.ParsedXML.NodePath import registry
from WidgetPathScheme import WidgetPathScheme
import WidgetRegistry
import EditorService

def initialize(context):
    context.registerClass(
        WidgetRegistry.WidgetRegistry,
        permission = "Add XMLWidget Registries",
        constructors = (WidgetRegistry.manage_addWidgetRegistryForm,
                        WidgetRegistry.manage_addWidgetRegistry),
        icon = "www/widget_service.gif"
        )
    
    context.registerClass(
        EditorService.EditorService,
        permission = "Add XMLWidget Registries",
        constructors = (EditorService.manage_addEditorServiceForm,
                        EditorService.manage_addEditorService),
        icon = "www/editor_service.gif"
        )

    registry.register_scheme(WidgetPathScheme('widget'))



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.11 $
"""
XMLWidgets support for DOM classes.
"""

import Globals
import Acquisition
from Products.ParsedXML import DOMProxy, DOM, ExtraDOM
from Products.ParsedXML.ManageableDOM import DOMTraversable
import xml.dom
from Products.ParsedXML.NodePath import registry

class WidgetAttribute(Acquisition.Implicit):
    
    def __getitem__(self, name):
        """
        """
        node = self.aq_parent
        return getattr(
            self.service_editor.getWidget(node), name)
    
   # def __getitem__(self, name):
   #     """
   #     """
   #     node = self.aq_parent
   #     wr = getattr(self, node.getWidgetRegistryName())
   #     return getattr(wr.getWidget(node), name)

def getWidgetWrappedNode(node):
    """Get a node for the widget wrapped with the WidgetDOM wrappers.
    """
    if node is None:
        return
    wrapper_type = WRAPPER_TYPES[node._get_nodeType()]
    parent = node.getPersistentDoc() or node
    return wrapper_type(node._node,
                        node._persistentDoc).__of__(parent)

class WidgetWrapper:
    """
    Mixin class to go alongside WidgetNode classes.
    """

    # anything that returns subobjs must grant access to them
    __allow_access_to_unprotected_subobjects__ = 1

    def wrapNamedNodeMap(self, obj):
        if obj is None:
            return None
        parent = self.getPersistentDoc() or self
        wrapped = WidgetNamedNodeMap(obj, self._persistentDoc).__of__(parent)
        return wrapped
        
    def wrapNodeList(self, obj):
        parent = self.getPersistentDoc() or self
        wrapped = WidgetNodeList(obj, self._persistentDoc).__of__(parent)      
        return wrapped

    def wrapDOMObj(self, node):
        """Return the appropriate class wrapped around the node."""
        if node is None:
            return
        wrapper_type = WRAPPER_TYPES[node._get_nodeType()]
        parent = self.getPersistentDoc() or self
        wrapped = wrapper_type(node,
                               self._persistentDoc).__of__(parent)      
        return wrapped
    
    def getWidgetRegistryName(self):
        session = self.REQUEST.SESSION
        # get document url XXX arrggh horrible hacks
        node = self
        while getattr(node.aq_base, 'meta_type', None) != 'Parsed XML':
            node = node.aq_parent
        document_url = node.absolute_url()
        document = node
        
        if session.has_key('xmlwidgets_node_event_handler'):
            xmlwidgets_node_event_handler = session['xmlwidgets_node_event_handler']
        else:
            xmlwidgets_node_event_handler = {}

        nodepath_wr_names = xmlwidgets_node_event_handler.get(
            document_url, None)
        if nodepath_wr_names:
            node = self
            while node is not None:
                path = registry.create_path(node.ownerDocument, node, 'widget')
                name = nodepath_wr_names.get(path, None)
                if name:
                    return name
                node = node.parentNode

        # if we can't find any special registry, we fall back on
        # document editor for now XXX
        if not session.has_key('xmlwidgets_event_handler'):
            session['xmlwidgets_event_handler'] = {}
        result = session['xmlwidgets_event_handler'].get(
            document_url, 'service_document_editor')
        return result

    def getWidgetRegistryNode(self):
        session = self.REQUEST.SESSION
        # get document url XXX arrggh horrible hacks
        node = self
        while getattr(node.aq_base, 'meta_type', None) != 'Parsed XML':
            node = node.aq_parent
        document_url = node.absolute_url()
        
        if session.has_key('xmlwidgets_node_event_handler'):
            xmlwidgets_node_event_handler = session['xmlwidgets_node_event_handler']
        else:
            xmlwidgets_node_event_handler = {}

        nodepath_wr_names = xmlwidgets_node_event_handler.get(
            document_url, None)
        if nodepath_wr_names:
            node = self
            while node is not None:
                path = registry.create_path(node.ownerDocument, node, 'widget')
                name = nodepath_wr_names.get(path, None)
                if name:
                    return node
                node = node.parentNode

        # if we can't find any special registry, return document element node
        return document.documentElement

    edit = WidgetAttribute()

class WidgetTraversable(DOMTraversable):
    """Traversing the widget way.
    """
    pass

    
# According to DOM Erratum Core-14, the empty string should be
# accepted as equivalent to null for hasFeature().

_WIDGET_DOM_FEATURES = (
    ("org.zope.dom.persistence", None),
    ("org.zope.dom.persistence", ""),
    ("org.zope.dom.persistence", "1.0"),
    ("org.zope.dom.acquisition", None),
    ("org.zope.dom.acquisition", ""),
    ("org.zope.dom.acquisition", "1.0"),
    )

_WIDGET_DOM_NON_FEATURES = (
    ("load", None),
    ("load", ""),
    ("load", "3.0"),
    )

class WidgetDOMImplementation(DOMProxy.DOMImplementationProxy):
    """A proxy of a DOMImplementation node that defines createDocument
    to return a WidgetDocument.
    """

    def hasFeature(self, feature, version):
        feature = string.lower(feature)
        if (feature, version) in _WIDGET_DOM_FEATURES:
            return 1
        if (feature, version) in _WIDGET_DOM_NON_FEATURES:
            return 0
        return DOMProxy.DOMImplementationProxy.hasFeature(self, feature,
                                                          version)

    def createDocumentType(self, qualifiedName, publicId, systemId):
        DOMDocumentType = self._createDOMDocumentType(qualifiedName,
                                                      publicId, systemId)
        return WidgetDocumentType(DOMDocumentType)
    
    def createDocument(self, namespaceURI, qualifiedName, docType=None):
        if docType is not None:
            if docType.ownerDocument is not None:
                raise xml.dom.WrongDocumentErr
            mdocType = docType.getDOMObj()
        else:
            mdocType = None
        DOMDocument = self._createDOMDocument(namespaceURI, qualifiedName,
                                              mdocType)
        return WidgetDocument(DOMDocument)

theDOMImplementation = WidgetDOMImplementation()
    
class WidgetNode(WidgetWrapper, DOMProxy.NodeProxy, WidgetTraversable,
                 Acquisition.Implicit):
    "A wrapper around a DOM Node."
    
class WidgetNodeList(WidgetWrapper, DOMProxy.NodeListProxy,
                     Acquisition.Implicit):
    "A wrapper around a DOM NodeList."
    meta_type = "Widget NodeList"
        
    # redefine to get back the [] syntax with acquisition, eh?
    def __getslice__(self, i, j):
        return self.wrapNodeList(self._node.__getslice__(i,j))

    # redefine to get back the [] syntax with acquisition, eh?
    def __getitem__(self, i):
        return self.wrapDOMObj(self._node.__getitem__(i))        

class WidgetNamedNodeMap(WidgetWrapper, DOMProxy.NamedNodeMapProxy,
                         Acquisition.Implicit):
    "A wrapper around a DOM NamedNodeMap."
    meta_type = "Widget NamedNodeMap"
        
    # redefine to get back the [] syntax with acquisition, eh?
    def __getitem__(self, i):
        return self.wrapDOMObj(self._node.__getitem__(i))        
    
class WidgetDocumentFragment(WidgetWrapper, DOMProxy.DocumentFragmentProxy,
                             WidgetNode):
    "A wrapper around a DOM DocumentFragment."
    meta_type = "Widget Document Fragment"
    
class WidgetElement(WidgetWrapper, DOMProxy.ElementProxy,
                    WidgetNode):
    "A wrapper around a DOM Element."
    meta_type = "Widget Element"

class WidgetCharacterData(WidgetWrapper,
                          DOMProxy.CharacterDataProxy, WidgetNode):
    "A wrapper around a DOM CharacterData."
    meta_type = "Widget Character Data"

class WidgetCDATASection(WidgetWrapper,
                         DOMProxy.CDATASectionProxy, WidgetNode):
    "A wrapper around a DOM CDATASection."
    meta_type = "Widget CDATASection"

class WidgetText(WidgetWrapper,
                 DOMProxy.TextProxy, WidgetCharacterData):
    "A wrapper around a DOM Text."
    meta_type = "Widget Text"

class WidgetComment(WidgetWrapper,
                    DOMProxy.CommentProxy, WidgetCharacterData):
    "A wrapper around a DOM Comment."
    meta_type = "Widget Comment"    

class WidgetProcessingInstruction(WidgetWrapper,
                                  DOMProxy.ProcessingInstructionProxy,
                                  WidgetNode):
    "A wrapper around a DOM ProcessingInstruction."
    meta_type = "Widget Processing Instruction"    

class WidgetAttr(WidgetWrapper, DOMProxy.AttrProxy, WidgetNode):
    "A wrapper around a DOM Attr."
    meta_type = "Widget Attr"    

#WidgetDocument is not necessarily a persistent object, even when a
#persistent subclass such as ParsedXML has been instantiated.  Traversing
#up to the document can create a new transient proxy.  Persistent attributes
#must be set on the persistent version.
class WidgetDocument(WidgetWrapper, DOMProxy.DocumentProxy,
                     WidgetNode):
    "A wrapper around a DOM Document."
    meta_type = "Widget Document"    

    implementation = theDOMImplementation

    def __init__(self, node, persistentDocument = None):
        WidgetNode.__init__(self, node, persistentDocument)

    def _get_implementation(self):
        return self.implementation
    
    #block set of implementation, since we don't proxy it the same
    def __setattr__(self, name, value):
        if name == "implementation":
            raise xml.dom.NoModificationAllowedErr()
        WidgetDocument.inheritedAttribute('__setattr__')(self, name, value)

# DOM extended interfaces

class WidgetEntityReference(WidgetWrapper,
                            DOMProxy.EntityReferenceProxy,
                            WidgetNode):
    "A wrapper around a DOM EntityReference."
    meta_type = "Widget Entity Reference"
    
class WidgetEntity(WidgetWrapper, DOMProxy.EntityProxy,
                   WidgetNode):
    "A wrapper around a DOM Entity."
    meta_type = "Widget Entity"

class WidgetNotation(WidgetWrapper, DOMProxy.NotationProxy,
                     WidgetNode):
    "A wrapper around a DOM Notation."
    meta_type = "Widget Notation"
    
class WidgetDocumentType(WidgetWrapper, DOMProxy.DocumentTypeProxy,
                         WidgetNode):
    "A wrapper around a DOM DocumentType."
    meta_type = "Widget Document Type"    


Node = xml.dom.Node
WRAPPER_TYPES = {
    Node.ELEMENT_NODE: WidgetElement,
    Node.ATTRIBUTE_NODE: WidgetAttr,
    Node.TEXT_NODE: WidgetText,
    Node.CDATA_SECTION_NODE: WidgetCDATASection,
    Node.ENTITY_REFERENCE_NODE: WidgetEntityReference,
    Node.ENTITY_NODE: WidgetEntity,
    Node.PROCESSING_INSTRUCTION_NODE: WidgetProcessingInstruction,
    Node.COMMENT_NODE: WidgetComment,
    Node.DOCUMENT_NODE: WidgetDocument,
    Node.DOCUMENT_TYPE_NODE: WidgetDocumentType,
    Node.DOCUMENT_FRAGMENT_NODE: WidgetDocumentFragment,
    Node.NOTATION_NODE: WidgetNotation,
    }
del Node



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.13 $
import unittest
try:
    import Zope2
except ImportError:  	# for Zope 2.7
    import Zope as Zope2
try:  			# Zope 2.8 style transactions
    import transaction
except ImportError:  	# Old-style transactions
    # BBB: for Zope 2.7
    class TransactionMan:
        def begin(self):                get_transaction().begin()
        def commit(self, sub=False):    get_transaction().commit(sub)
        def abort(self, sub=False):     get_transaction().abort(sub)
        def get(self):                  return get_transaction()
    transaction = TransactionMan()

from Products.ParsedXML.ParsedXML import ParsedXML

xml = '''\
<?xml version="1.0" ?>
<doc><p>A paragraph</p><p>Paragraph two, the sequel.</p><section><p>Sub p</p><section><p>Sub sub</p></section></section></doc>
'''
class FakeRequest:
    def __init__(self):
        self.SESSION = {}
        self.dict = { 'SERVER_URL' : 'this_server'}
        self._script = ["script"]
        
    def __getitem__(self, key):
        return self.dict[key]
    
    def __setitem__(self, key, object):
        self.dict[key] = object
    
    def get(self, key, default):
        return self.dict.get(key, default)

class EditorServiceTestCase(unittest.TestCase):
    def setUp(self):
        transaction.begin()
        self.connection = Zope2.DB.open()
        self.root = root = self.connection.root()['Application']
        # create documents
        root.manage_addProduct['ParsedXML'].manage_addParsedXML(
            'doc1', file=xml)
        root.manage_addProduct['ParsedXML'].manage_addParsedXML(
            'doc2', file=xml)
        root.REQUEST = FakeRequest()
        self.doc1 = root.doc1
        self.doc2 = root.doc2
        # create editor service
        root.manage_addProduct['XMLWidgets'].manage_addEditorService(
            'service_editor')
        self.s = root.service_editor
        # create widget folders
        root.manage_addFolder('alpha')
        alpha = root.alpha
        alpha.manage_addFolder('doc')
        alpha.manage_addFolder('p')
        alpha.manage_addFolder('section')
        
        root.manage_addFolder('beta')
        beta = root.beta
        beta.manage_addFolder('doc')
        beta.manage_addFolder('section')
        beta.manage_addFolder('p')
        beta.manage_addFolder('section_focus')
        beta.manage_addFolder('p_focus')

        root.manage_addFolder('gamma')
        gamma = root.gamma
        gamma.manage_addFolder('section')
        gamma.manage_addFolder('p')
        gamma.manage_addFolder('section_focus')
        gamma.manage_addFolder('p_focus')

        # create and configure widget registries
        root.manage_addProduct['XMLWidgets'].manage_addWidgetRegistry(
            'wr1')
        self.wr1 = root.wr1
        root.manage_addProduct['XMLWidgets'].manage_addWidgetRegistry(
            'wr2')
        self.wr2 = root.wr2
        root.manage_addProduct['XMLWidgets'].manage_addWidgetRegistry(
            'wr3')
        self.wr3 = root.wr3
        
        self.wr1.addWidget('doc', ('alpha', 'doc'))
        self.wr1.addWidget('p', ('alpha', 'p'))
        self.wr1.addWidget('section', ('alpha', 'section'))

        self.wr2.addWidget('doc', ('beta', 'doc'))
        self.wr2.addWidget('section', ('beta', 'section'))
        self.wr2.addWidget('p', ('beta', 'p'))

        self.wr3.addWidget('section', ('gamma', 'section'))
        self.wr3.addWidget('p', ('gamma', 'p'))
        
    def tearDown(self):
        transaction.abort()
        self.connection.close()

    def test_DocumentEditor(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        self.assertEquals(
            ('alpha', 'doc'),
            self.s._getWidgetPath(self.doc1.documentElement))
        self.assertEquals(
            self.root.alpha.doc,
            self.s.getWidget(
            self.doc1.documentElement))
        node = self.doc1.documentElement.childNodes[0]
        self.assertEquals(
            self.root.alpha.p,
            self.s.getWidget(node))
        node = self.doc1.documentElement.childNodes[2]
        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(node))
        
    def test_DocumentEditor2(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        self.s.setDocumentEditor(self.doc2, 'wr2')
        self.assertEquals(
            self.root.alpha.doc,
            self.s.getWidget(
            self.doc1.documentElement))
        self.assertEquals(
            self.root.beta.doc,
            self.s.getWidget(
            self.doc2.documentElement))  
        
    def test_Editor(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        self.s.setEditor(
            self.doc1.documentElement.childNodes[2],
            'wr2')
        self.assertEquals(
            self.root.alpha.doc,
            self.s.getWidget(
            self.doc1.documentElement))
        self.assertEquals(
            self.root.alpha.p,
            self.s.getWidget(
            self.doc1.documentElement.childNodes[1]))
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(
            self.doc1.documentElement.childNodes[2]))
        self.assertEquals(
            self.root.beta.p,
            self.s.getWidget(
            self.doc1.documentElement.childNodes[2].childNodes[0]))

    def test_NodeWidget(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        section_node = self.doc1.documentElement.childNodes[2]
        self.s.setNodeWidget(section_node,
                             ('beta', 'section'))
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(section_node))

        p_node = self.doc1.documentElement.childNodes[0]
        self.s.setNodeWidget(p_node,
                             ('beta', 'p'))
        self.assertEquals(
            self.root.beta.p,
            self.s.getWidget(p_node))

        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))

    def test_NodeWidget2(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        section_node = self.doc1.documentElement.childNodes[2]
        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))
        
        self.s.setNodeWidget(section_node,
                             ('beta', 'section'))
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(section_node))

        self.s.clearNodeWidget(self.doc1.documentElement)
 
        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))

    def test_EditorNodeWidget(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        section_node = self.doc1.documentElement.childNodes[2]
        self.s.setNodeWidget(section_node,
                             ('beta', 'section'))
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(section_node))

        p_node = self.doc1.documentElement.childNodes[0]
        self.s.setNodeWidget(p_node,
                             ('beta', 'p'))
        self.assertEquals(
            self.root.beta.p,
            self.s.getWidget(p_node))

        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))

    def test_EditorNodeWidget2(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        section_node = self.doc1.documentElement.childNodes[2]
        self.s.setNodeWidget(section_node,
                             ('beta', 'section'))
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(section_node))

        p_node = self.doc1.documentElement.childNodes[0]
        self.s.setNodeWidget(p_node,
                             ('beta', 'p'))
        self.assertEquals(
            self.root.beta.p,
            self.s.getWidget(p_node))

        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))
        
    def test_Root(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        section_node = self.doc1.documentElement.childNodes[2]
        self.s.pushRoot(section_node, 'wr2')
        self.assertEquals(
            self.root.beta.section,
            self.s.getWidget(section_node))
        self.s.popRoot(section_node)
        self.assertEquals(
            self.root.alpha.section,
            self.s.getWidget(section_node))
        
    def test_getNode(self):
        self.s.setDocumentEditor(self.doc1, 'wr1')
        p_node = self.doc1.documentElement.childNodes[0]
        section_node = self.doc1.documentElement.childNodes[2]
        doc_node = self.doc1.documentElement
        self.s.setNodeWidget(
            section_node, ('beta', 'section'))
        self.assertEquals(
            section_node, self.s.getNode(doc_node))
        self.s.setNodeWidget(
            p_node, ('beta', 'p'))
        self.assertEquals(
            p_node, self.s.getNode(doc_node))
        # Clear the node widget, thus getNode should return None
        self.s.clearNodeWidget(p_node)
        self.assertEquals(
            None, self.s.getNode(doc_node))

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(EditorServiceTestCase, 'test'))
    return suite

def main():
    unittest.TextTestRunner().run(test_suite())

if __name__ == '__main__':
    main()



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.9 $
import unittest
try:
    import Zope2
except ImportError:  	# for Zope 2.7
    import Zope as Zope2
try:
    startup = Zope2.startup
except (AttributeError, ImportError):
    # startup only exists in Zope 2.6+
    pass
else:
    startup()

from Products.XMLWidgets.tests import test_EditorService
from Products.XMLWidgets.tests import test_EditorCache

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(test_EditorService.test_suite())
    suite.addTest(test_EditorCache.test_suite())
    return suite

def main():
    unittest.TextTestRunner(verbosity=1).run(test_suite())

if __name__ == '__main__':
    main()



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.22 $
from AccessControl import ClassSecurityInfo
import Globals
import zLOG
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from helpers import add_and_edit

class WidgetRegistry(SimpleItem):
    security = ClassSecurityInfo()

    meta_type = "XMLWidgets Registry"

    manage_options = (
        {'label':'Info', 'action':'info_tab'},
        ) + SimpleItem.manage_options

    info_tab = PageTemplateFile(
        'www/widgetRegistryInfoTab',
        globals(),  __name__='info_tab')

    # needed, as this uses here/manage_page_header
    # which need the "manager" role
    security.declareProtected('View management screens', 'info_tab')

    def __init__(self, id, title):
        self.id = id
        self.title = title
        self._node_map = {}
        self._allowed_map = {}
        self._display_map = {}
        
    security.declareProtected('Change XMLWidgets Registries',
                              'addWidget')
    def addWidget(self, nodeName, widget_path):
        self._node_map[nodeName] = widget_path
        self._node_map = self._node_map

    security.declareProtected('Change XMLWidgets Registries',
                              'clearWidgets')
    def clearWidgets(self):
        """Clear all widgets.
        """
        self._node_map = {}
        self._allowed_map = {}
        self._display_map = {}

    security.declareProtected('Change XMLWidgets Registries',
                              'setAllowed')
    def setAllowed(self, nodeName, allowed):
        """Set which names can be inserted in another node.
        """
        self._allowed_map[nodeName] = allowed
        self._allowed_map = self._allowed_map

    security.declareProtected('Change XMLWidgets Registries',
                              'setDisplayName')
    def setDisplayName(self, nodeName, displayName):
        """Set visible name for a nodeName.
        """
        self._display_map[nodeName] = displayName
        self._display_map = self._display_map
    
    security.declareProtected('View', 'findWidget')
    def findWidget(self, node):
        widget_path = self._node_map.get(node.nodeName, None)
        if widget_path is None:
            zLOG.LOG('XMLWidgets',zLOG.ERROR,
                     '%s: no view widget registered for node %s' % \
                     (self.getId(), node.nodeName))
            return None        
        return widget_path

    security.declareProtected('View', 'getWidget')
    def getWidget(self, node):
        widget_path = self._node_map.get(node.nodeName, None)
        if widget_path is None:
            raise AttributeError, ('%s: no widget registered for node name %s' %\
                                   (self.getId(), node.nodeName))
        obj = self.aq_inner
        for step in widget_path:
            obj = getattr(obj, step, None)
            if obj is None:
                raise AttributeError, \
                      ('XML Widget registry %s cannot acquire %s ("%s" not found)' % \
                       ( self.getId(), '/'.join(widget_path), step ))
        self.REQUEST['node'] = node
        return obj
    
    security.declareProtected('View', 'getDisplayName')
    def getDisplayName(self, nodeName):
        """Get the list of nodes allowed in nodeName.
        """
        return self._display_map.get(nodeName, nodeName)
    
    security.declareProtected('View', 'isAllowed')
    def isAllowed(self, parentName, name):
        """Check whether we are allowed in parentName.
        """
        return name in self._allowed_map.get(parentName, [])

    security.declareProtected('View', 'getAllowed')
    def getAllowed(self, parentName):
        """Give list of what is allowed in this parent.
        """
        return self._allowed_map.get(parentName, [])
    
Globals.InitializeClass(WidgetRegistry)

manage_addWidgetRegistryForm = PageTemplateFile(
    "www/widgetRegistryAdd", globals(),
    __name__='manage_addWidgetRegistryForm')

def manage_addWidgetRegistry(self, id, title="", REQUEST=None):
    """Add widget registry to folder.
    """
    # add actual object
    id = self._setObject(id, WidgetRegistry(id, title))
    # respond to the add_and_edit button if necessary
    add_and_edit(self, id, REQUEST)
    return ''



import unittest
from Products.XMLWidgets import EditorService
from Products.XMLWidgets.EditorService import EditorCache, PersistentEditorCache

class EditorCacheTestCase(unittest.TestCase):
    def setUp(self):
        self.cache = EditorCache()

    def test_cache(self):
        self.cache.cache(1, 1, 1, 'foo')
        self.assertEquals('foo', self.cache.getCached(1, 1, 1))

    def test_notCached(self):
        self.assertEquals(None, self.cache.getCached(1, 1, 1))
        
    def test_invalidate(self):
        self.cache.cache(1, 1, 1, 'foo')
        self.cache.invalidate(1, 1, 1)
        self.assertEquals(None, self.cache.getCached(1, 1, 1))

    def test_clearDocument(self):
        self.cache.cache(1, 1, 1, 'foo')
        self.cache.cache(1, 2, 3, 'bar')
        self.cache.clearDocument(1)
        self.assertEquals(None, self.cache.getCached(1, 1, 1))
        self.assertEquals(None, self.cache.getCached(1, 2, 3))

    def test_clear(self):
        self.cache.cache(1, 1, 1, 'foo')
        self.cache.cache(1, 2, 3, 'bar')
        self.cache.cache(2, 1, 1, 'baz')
        self.cache.clear()
        self.assertEquals(None, self.cache.getCached(1, 1, 1))
        self.assertEquals(None, self.cache.getCached(1, 2, 3))
        self.assertEquals(None, self.cache.getCached(2, 1, 1))

    def test_setParameters(self):
        self.assertEquals(EditorService.CACHE_SIZE, self.cache._cache_size)
        new_size = 1234
        self.cache.setCacheSize(new_size)
        self.assertEquals(new_size, self.cache._cache_size)

    def test_compact(self):        
        self.cache.cache(1, 1, 1, 'foo')
        self.cache.cache(2, 2, 3, 'bar')
        self.cache.cache(3, 1, 1, 'baz')
        self.cache.cache(4, 1, 1, 'kermit')
        self.cache.cache(5, 2, 3, 'ms. piggy')
        self.cache.cache(6, 1, 1, 'gonzo')
        self.assertEquals(6, len(self.cache._cache))
        # nothing should change.
        self.cache.setCacheSize(10)
        self.cache.compact()        
        self.assertEquals(6, len(self.cache._cache))
        # cache is larger then new size, so it should compact.
        self.cache.setCacheSize(4)
        self.cache.compact()
        self.assertEquals(2, len(self.cache._cache))

class PersistentEditorCacheTestCase(EditorCacheTestCase):
    def setUp(self):
        self.cache = PersistentEditorCache()
    
def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(EditorCacheTestCase, 'test'))
    suite.addTest(unittest.makeSuite(PersistentEditorCacheTestCase, 'test'))
    return suite

def main():
    unittest.TextTestRunner().run(test_suite())

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import os

version = '0.13.1'

setup(name='Products.XMLWidgets',
      version=version,
      description="XMLWidgets is a Zope 2 product used to create through the web viewers and editors of ParsedXML content.",
      long_description=open(os.path.join("Products", "XMLWidgets", "README.txt")).read() + "\n" +
                       open(os.path.join("Products", "XMLWidgets", "CREDITS.txt")).read() + "\n" +
                       open(os.path.join("Products", "XMLWidgets", "HISTORY.txt")).read(),
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
              "Framework :: Zope2",
              "License :: OSI Approved :: BSD License",
              "Programming Language :: Python",
              "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
              "Topic :: Software Development :: Libraries :: Python Modules",
              "Topic :: Software Development :: Widget Sets",
              "Topic :: Text Processing :: Markup :: HTML",
              "Topic :: Text Processing :: Markup :: XML",
              ],
      keywords='xml rendering zope2 parsedxml',
      author='Martijn Faassen',
      author_email='info@infrae.com',
      url='http://infrae.com/products/silva',
      license='BSD',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
              'setuptools',
              'Products.ParsedXML'
              ],
      )



# Copyright (c) 2002-2005 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.39 $
import Globals
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from BTrees.OOBTree import OOBTree
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from helpers import add_and_edit
from Persistence import Persistent

CACHE_SIZE = 1000 # max 1000 documents in cache, if more, shrink
CACHE_SHRINK_FACTOR = 2 # shrink cache by which factor if over size

theEditorCache = None

class EditorServiceError(Exception):
    pass

class EditorCache:
    """Rendered widgets cache for EditorService
    """

    def __init__(self,
                 cache_size=CACHE_SIZE,
                 cache_shrink_factor=CACHE_SHRINK_FACTOR):
        self._cache_size = cache_size
        self._cache_shrink_factor = cache_shrink_factor
        self._cache = self._new_mapping()
        
    # MANIPULATORS

    def _new_mapping(self):
        return {}

    def cache(self, document_key, node_path, widget_path, data):
        # compact cache if necessary
        self.compact()
        # now add this to cache
        if not self._cache.has_key(document_key):
            self._cache[document_key] = self._new_mapping()
        self._cache[document_key][(node_path, widget_path)] = data
    
    def invalidate(self, document_key, node_path, widget_path):
        try:
            del self._cache[document_key][(node_path, widget_path)]
        except KeyError:
            pass
        
    def clearDocument(self, document_key):
        try:
            del self._cache[document_key]
        except KeyError:
            pass

    def clear(self):
        self._cache = self._new_mapping()

    def compact(self):
        """Shrink cache is necessary.

        Note that cache shrinking is currently arbitrary, but most used
        documents will appear in the cache again soon enough at only a
        minor penalty of performance, so the simplicity is worth it
        """
        if len(self._cache) < self._cache_size:
            return

        shrinked_size = int(self._cache_size / self._cache_shrink_factor)        
        for i in range(len(self._cache) - shrinked_size):
            # pop items from dict until dict length equals shrinked_size
            self._cache.popitem()

    def setCacheSize(self, size):
        self._cache_size = size

    # ACCESSORS

    def getCached(self, document_key, node_path, widget_path):
        try:
            return self._cache[document_key][(node_path, widget_path)]
        except KeyError:
            return None

class PersistentEditorCache(Persistent, EditorCache):
    """Rendered widgets cache for EditorService

    This cache can be attached to e.g. a Zope folder. This Folder could then 
    be a mount point on a ZEO Storage Server, effectively sharing this cache
    over the ZEO Cluster.
    """

    def _new_mapping(self):
        return OOBTree()

    def compact(self):
        if len(self._cache) < self._cache_size:
            return

        shrinked_size = int(self._cache_size / self._cache_shrink_factor)        
        for i in range(len(self._cache) - shrinked_size):
            # delete items from btree until btree length equals shrinked_size
            key = self._cache.minKey()
            del self._cache[key]

class EditorSession(Persistent):
    def __init__(self):
        self.document_context = {
            'depth': 0,
            'wr_name': None,
            'node_path': None,
            'widget_path': None,
            }
        self.contexts = {}
        self.roots = []

    # MANIPULATORS
    def setDocumentContext(self, wr_name):
        self.document_context['wr_name'] = wr_name
        self.document_context = self.document_context
        
    def setContext(self, node, wr_name):
        self.clearNodeWidget(node)
        self.contexts[node.getNodePath('widget')] = {
            'depth': self.getContext(node)['depth'] + 1,
            'wr_name': wr_name,
            'node_path': None,
            'widget_path': None
            }
        self.contexts = self.contexts
        
    def setNodeWidget(self, node, widget_path):
        self.clearNodeWidget(node)
        context = self.getContext(node)
        context['node_path'] = node.getNodePath('widget')
        context['widget_path'] = widget_path
        self.document_context = self.document_context
        self.contexts = self.contexts
        
    def clearNodeWidget(self, node):
        context = self.getContext(node)
        context['node_path'] = None
        context['widget_path'] = None
        d = context['depth']
        for node_path, context in self.contexts.items():
            if context['depth'] > d:
                del self.contexts[node_path]
        self.contexts = self.contexts
                
    def clearContext(self, node):
        while node is not None:
            node_path = node.getNodePath('widget')
            context = self.contexts.get(node_path)
            if context is not None:
                del self.contexts[node_path]
                self.contexts = self.contexts
                return
            node = node.parentNode
    
    def pushRoot(self, node, wr_name):
        self.setContext(node, wr_name)
        self.roots.append(node.getNodePath('widget'))
        self.roots = self.roots

    def popRoot(self, node):
        self.clearContext(node)
        self.roots.pop()
        self.roots = self.roots
        
    # ACCESSORS
    def getContext(self, node):
        if not self.contexts:
            return self.document_context
        while node is not None:
            context = self.contexts.get(node.getNodePath('widget'))
            if context is not None:
                return context
            node = node.parentNode
        return self.document_context

    def getRoot(self):
        if self.roots:
            return self.roots[-1]
        else:
            return None
        
class EditorService(SimpleItem):
    """A user using an editor uses the editor session (which uses
    Zope's sessions).
    """

    security = ClassSecurityInfo()
    
    meta_type = "XMLWidgets Editor Service"

    cache_container_id = None

    cache_size = CACHE_SIZE
    cache_shrink_factor = CACHE_SHRINK_FACTOR
    
    manage_options = (
        {'label':'Info', 'action':'info_tab'},
        {'label':'Edit', 'action':'manage_editorServiceCacheEditTab'},
        ) + SimpleItem.manage_options

    security.declareProtected(
        'View management screens', 'manage_editorServiceCacheEditTab')
    manage_editorServiceCacheEditTab = PageTemplateFile(
        'www/editorServiceEditTab', globals(), 
        __name__='manage_editorServiceCacheEditTab')

    security.declareProtected(
        'View management screens', 'info_tab')
    manage_main = info_tab = PageTemplateFile(
        'www/editorServiceInfoTab', globals(),  __name__='info_tab')

    def __init__(self, id, title):
        self.id = id
        self.title = title

    # MANIPULATORS
    security.declareProtected(
        'View management screens', 'manage_editorServiceCacheEditTab')
    def manage_editorServiceCacheEdit(self, id, size):
        """set the cache container
        """
        msg = 'Settings changed.'

        if not id:
            # id left empty, use None - use in-memory cache
            id = None
            msg += ' Id left empty, so using in-memory cache.'        
        elif getattr(self.aq_inner, id, None) is None:
            # not a valid object, report that..
            return self.manage_editorServiceCacheEditTab(
                manage_tabs_message='Id does not provide for an existing cache container.')

        # Set parameters.
        self.cache_container_id = id
        self.cache_size = size  
        # Set it on the live cache object too, and compact it.
        cache = self._get_editor_cache()
        cache.setCacheSize(size)
        cache.compact()

        return self.manage_editorServiceCacheEditTab(manage_tabs_message=msg)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'setDocumentEditor')
    def setDocumentEditor(self, node, wr_name):
        self._createEditorSession(node)
        self._getEditorSession(node).setDocumentContext(wr_name)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'setEditor')    
    def setEditor(self, node, wr_name):
        self._getEditorSession(node).setContext(node, wr_name)

    security.declareProtected('View',
                              'setViewer')
    def setViewer(self, wr_name):
        self.REQUEST.set('xmlwidgets_viewer', wr_name)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearEditor')
    def clearEditor(self, node):
        self._getEditorSession(node).clearContext(node)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'setNodeWidget')
    def setNodeWidget(self, node, widget_path):
        self._getEditorSession(node).setNodeWidget(node, widget_path)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearNodeWidget') 
    def clearNodeWidget(self, node):
        self._getEditorSession(node).clearNodeWidget(node)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'pushRoot')
    def pushRoot(self, node, wr_name):
        self._getEditorSession(node).pushRoot(node, wr_name)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'popRoot')
    def popRoot(self, node):
        self._getEditorSession(node).popRoot(node)

    # ACCESSORS

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getRoot')
    def getRoot(self, node):
        document = node.ownerDocument
        editor_session = self._getEditorSession(node)
        if editor_session is None:
            return document.documentElement
        root = editor_session.getRoot()
        if root is not None:
            return document.resolveNodePath(root)
        else:
            return document.documentElement

    security.declareProtected('View', 'getViewer')
    def getViewer(self):
        return getattr(self, self.REQUEST['xmlwidgets_viewer'])

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getNode')
    def getNode(self, node):
        """Get the highlighted node in context.
        """
        node_path = self._getEditorSession(node).getContext(None)['node_path']
        if node_path is not None:
            to_ret = node.ownerDocument.resolveNodePath(node_path)
            return to_ret
        else:
            return None
        
    security.declareProtected('View', 'renderView')
    def renderView(self, node):
        # NOTE: Need to store viewer as it may change otherwise
        # during rendering XXX does this still make sense?
        viewer = self.getViewer()
        return viewer.getWidget(node).render()

    security.declareProtected('View', 'renderElementsView')
    def renderElementsView(self, node):
        # NOTE: Need to store viewer as it may change otherwise
        # during rendering XXX does this still make sense?
        viewer = self.getViewer()
        return ''.join([viewer.getWidget(child).render()
                        for child in node.childNodes
                        if child.nodeType == node.ELEMENT_NODE])

    security.declareProtected('Use XMLWidgets Editor Service',
                              'render')
    def render(self, node):
        document_key = self._getDocumentKey(node)
        node_path = node.getNodePath('widget')
        context = self._getEditorSession(node).getContext(node)
        widget_path = self._getWidgetPathInContext(context, node)
        request = self.REQUEST
        request['node'] = node
        request.wr_name = context['wr_name']
        cache = self._get_editor_cache()
        data = cache.getCached(document_key, node_path, widget_path)
        if data is None:
            widget = self._resolveWidgetPath(widget_path)
            __traceback_info__ = ('/'.join(widget_path), widget)
            data = widget.render()
            if getattr(widget, 'cached', lambda: 0)():
#                print "Caching:", widget_path, node_path
                cache.cache(
                    document_key, node_path, widget_path, data)
#            else:
#                print "Not caching:", widget_path, node_path
#        else:
#            print "Retrieving:", widget_path, node_path
        return data

    # XXX renderCache deprecated
    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderCache')
    renderCache = render
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderElements')
    def renderElements(self, node):
        return ''.join([self.render(child)
                        for child in node.childNodes if
                        child.nodeType == node.ELEMENT_NODE])

    # XXX renderElementsCache deprecated
    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderElementsCache')
    renderElementsCache = renderElements
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateWidgetCache')
    def invalidateWidgetCache(self, node, widget_path):
        """Invalidates cache for 'modeless' widgets"""
        document_key = self._getDocumentKey(node)
        cache = self._get_editor_cache()
        cache.invalidate(
            document_key,
            node.getNodePath('widget'),
            widget_path)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateWidgetCaches')
    def invalidateWidgetCaches(self, node, widget_path, modes):
        document_key = self._getDocumentKey(node)
        node_path = node.getNodePath('widget')
        editor_session = self._getEditorSession(node)
        cache = self._get_editor_cache()
        for mode in modes:            
            cache.invalidate(
                document_key,
                node_path,
                widget_path + (mode,))
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateCaches')
    def invalidateCaches(self, node, modes):
        widget_path = self._getWidgetPath(node)[:-1]
        document_key = self._getDocumentKey(node)
        node_path = node.getNodePath('widget')
        editor_session = self._getEditorSession(node)
        cache = self._get_editor_cache()
        for mode in modes:            
            cache.invalidate(
                document_key,
                node_path,
                widget_path + (mode,))
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateChildCaches')
    def invalidateChildCaches(self, node, modes):
        """Invalidate cache of children (one level)
        """
        for child in node.childNodes:
            if child.nodeType != node.ELEMENT_NODE:
                continue
            self.invalidateCaches(child, modes)
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearCache')    
    def clearCache(self, node):
        cache = self._get_editor_cache()
        cache.clearDocument(self._getDocumentKey(node))

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getWidget')
    def getWidget(self, node):
        context = self._getEditorSession(node).getContext(node)
        self.REQUEST['node'] = node
        self.REQUEST.wr_name = context['wr_name']
        widget_path = self._getWidgetPathInContext(context, node)
        if widget_path is None:
            raise EditorServiceError, "Unknown widget for node: %s" % node.nodeName
        return self._resolveWidgetPath(widget_path)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'hasNodeWidget')
    def hasNodeWidget(self, node):
        context = self._getEditorSession(node).getContext(node)
        return context['node_path'] is not None
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'getNodeWidgetPath')
    def getNodeWidgetPath(self, node):
        editor_session = self._getEditorSession(node)
        if not editor_session:
            return None
        return editor_session.getContext(node)['widget_path']
    
    # HELPERS

    def _getWidgetPathInContext(self, context, node):
        if (context['node_path'] and
            node.getNodePath('widget') == context['node_path']):
            return context['widget_path']
        else:
            return getattr(self, context['wr_name']).findWidget(node)
        
    def _getWidgetPath(self, node):
        context = self._getEditorSession(node).getContext(node)
        return self._getWidgetPathInContext(context, node)
        
    def _resolveWidgetPath(self, widget_path):
        obj = self.aq_inner
        for step in widget_path:
            obj = getattr(obj, step, None)
            if obj is None:
                raise AttributeError, \
                      ( 'XML Editor Service %s cannot acquire %s ("%s" not found)' % \
                        ( self.getId(), '/'.join(widget_path), step ) )
        return obj

    def _createEditorSession(self, node):
        session = self.REQUEST.SESSION
        session_data = session.get('xmlwidgets_service_editor', {})
        document_key =  self._getDocumentKey(node)
        if session_data.has_key(document_key):
            return
        editor_session = session_data[document_key] = EditorSession()
        session['xmlwidgets_service_editor'] = session_data
        
    def _getEditorSession(self, node):
        try:
            document_key = self._getDocumentKey(node)
            return self.REQUEST.SESSION[
                'xmlwidgets_service_editor'][document_key]
        except KeyError:
            return None

    def _getDocumentKey(self, node):
        if node.nodeType == node.DOCUMENT_NODE: 	 
            doc = node 	 
        else: 	 
            doc = node.ownerDocument
            
        request = self.REQUEST        
        path =  doc.getPhysicalPath()
        return (request['SERVER_URL'], tuple(request._script), path)        
        
    def _get_editor_cache(self):
        if self.cache_container_id is None:
            # Do not attach cache to a container - use in-memory cache
            # using theEditorCache singleton.
            global theEditorCache
            if theEditorCache is None:
                theEditorCache = EditorCache(
                    cache_size=self.cache_size,
                    cache_shrink_factor=self.cache_shrink_factor)                
            return theEditorCache
        else:
            # Attach cache to a container - useful for sharing the cache
            # over ZEO clients (this container should then be mounted
            # from the ZEO Storage Server).
            cc = getattr(self.aq_inner, self.cache_container_id)
            storage = getattr(cc, 'widget_editor_cache', None)
            if storage is None:
                cc.widget_editor_cache = PersistentEditorCache(
                    cache_size=self.cache_size,
                    cache_shrink_factor=self.cache_shrink_factor)
                # Trigger persistence machinery
                cc._p_changed = 1
                storage = cc.widget_editor_cache
            return storage
        
Globals.InitializeClass(EditorService)

manage_addEditorServiceForm = PageTemplateFile(
    "www/editorServiceAdd", globals(),
    __name__='manage_addEditorServiceForm')

def manage_addEditorService(self, id, title="", REQUEST=None):
    """Add Editor service to folder.
    """
    # add actual object
    id = self._setObject(id, EditorService(id, title))
    # respond to the add_and_edit button if necessary
    add_and_edit(self, id, REQUEST)
    return ''
